import os
from typing import Any, Dict, Optional

from pydantic import BaseModel
import requests

from dbt_loom.logging import fire_event


class DbtCloudReferenceConfig(BaseModel):
    """Configuration for a dbt Cloud reference."""

    account_id: int
    job_id: int
    api_endpoint: Optional[str] = None
    step: Optional[int] = None


class DbtCloud:
    """API Client for dbt Cloud. Fetches latest manifest for a given dbt job."""

    def __init__(
        self,
        account_id: int,
        token: Optional[str] = None,
        api_endpoint: Optional[str] = None,
    ) -> None:
        resolved_token = token or os.environ.get("DBT_CLOUD_API_TOKEN")
        if resolved_token is None:
            raise Exception(
                "A DBT Cloud token must be provided to dbt-loom when fetching manifest "
                "data from dbt Cloud. Please provide one via the `DBT_CLOUD_API_TOKEN` "
                "environment variable."
            )

        self.__token: str = resolved_token

        self.account_id = account_id
        self.api_endpoint = api_endpoint or "https://cloud.getdbt.com/api/v2"

    def _query(self, endpoint: str, **kwargs) -> Dict:
        """Query the dbt Cloud Administrative API."""
        url = f"{self.api_endpoint}/{endpoint}"
        fire_event(msg=f"Querying {url}")
        response = requests.get(
            url,
            headers={
                "authorization": "Bearer " + self.__token,
                "content-type": "application/json",
            },
            **kwargs,
        )
        return response.json()

    def _get_manifest(self, run_id: int, step: Optional[int] = None) -> Dict[str, Any]:
        """Get the manifest json for a given dbt Cloud run."""
        params = {}
        if step:
            params["step"] = step

        return self._query(
            f"accounts/{self.account_id}/runs/{run_id}/artifacts/manifest.json",
            params=params,
        )

    def _get_latest_run(self, job_id: int) -> Dict[str, Any]:
        """Get the latest run performed by a dbt Cloud job."""
        return self._query(
            f"accounts/{self.account_id}/runs/",
            params={
                "job_definition_id": job_id,
                "status": 10,
                "order_by": "-finished_at",
                "limit": 1,
            },
        )["data"][0]

    def get_models(self, job_id: int, step: Optional[int] = None) -> Dict[str, Any]:
        """Get the latest state of all models by Job ID."""
        latest_run = self._get_latest_run(job_id=job_id)
        return self._get_manifest(run_id=latest_run["id"], step=step)



import json
from pathlib import Path
from typing import Dict, Optional

import boto3
import gzip
from io import BytesIO
from pydantic import BaseModel


class S3ReferenceConfig(BaseModel):
    """Configuration for an reference stored in S3"""

    bucket_name: str
    object_name: str
    credentials: Optional[Path] = None


class S3Client:
    """A client for loading manifest files from S3-compatible object stores."""

    def __init__(self, bucket_name: str, object_name: str) -> None:
        self.bucket_name = bucket_name
        self.object_name = object_name

    def load_manifest(self) -> Dict:
        """Load the manifest.json file from an S3 bucket."""

        client = boto3.client("s3")

        # TODO: Determine if I need to add args for SSE
        try:
            response = client.get_object(Bucket=self.bucket_name, Key=self.object_name)
        except client.exceptions.NoSuchBucket:
            raise Exception(f"The bucket `{self.bucket_name}` does not exist.")
        except client.exceptions.NoSuchKey:
            raise Exception(
                f"The object `{self.object_name}` does not exist in bucket "
                f"`{self.bucket_name}`."
            )

        # Deserialize the body of the object.
        try:
            if self.object_name.endswith(".gz"):
                body = response["Body"].read()
                with gzip.GzipFile(fileobj=BytesIO(body)) as gzipfile:
                    content = gzipfile.read().decode('utf-8')
            else:
                content = response["Body"].read().decode("utf-8")
        except Exception:
            raise Exception(
                f"Unable to read the data contained in the object `{self.object_name}"
            )

        try:
            return json.loads(content)
        except Exception:
            raise Exception(
                f"The object `{self.object_name}` does not contain valid JSON."
            )



import datetime
from io import BytesIO
import json
import gzip
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import urlunparse

from pydantic import BaseModel, Field, validator
import requests

try:
    from dbt.artifacts.resources.types import NodeType
except ModuleNotFoundError:
    from dbt.node_types import NodeType  # type: ignore

from dbt_loom.clients.az_blob import AzureClient, AzureReferenceConfig
from dbt_loom.clients.dbt_cloud import DbtCloud, DbtCloudReferenceConfig
from dbt_loom.clients.gcs import GCSClient, GCSReferenceConfig
from dbt_loom.clients.s3 import S3Client, S3ReferenceConfig
from dbt_loom.config import (
    FileReferenceConfig,
    LoomConfigurationError,
    ManifestReference,
    ManifestReferenceType,
)


class DependsOn(BaseModel):
    """Wrapper for storing dependencies"""

    nodes: List[str] = Field(default_factory=list)
    macros: List[str] = Field(default_factory=list)


class ManifestNode(BaseModel):
    """A basic ManifestNode that can be referenced across projects."""

    name: str
    package_name: str
    unique_id: str
    resource_type: NodeType
    schema_name: str = Field(alias="schema")
    database: Optional[str] = None
    relation_name: Optional[str] = None
    version: Optional[str] = None
    latest_version: Optional[str] = None
    deprecation_date: Optional[datetime.datetime] = None
    access: Optional[str] = "protected"
    group: Optional[str] = None
    generated_at: datetime.datetime = Field(default_factory=datetime.datetime.utcnow)
    depends_on: Optional[DependsOn] = None
    depends_on_nodes: List[str] = Field(default_factory=list)
    enabled: bool = True

    @validator("depends_on_nodes", always=True)
    def default_depends_on_nodes(cls, v, values):
        depends_on = values.get("depends_on")
        if depends_on is None:
            return []

        return [
            node for node in depends_on.nodes if node.split(".")[0] not in ("source")
        ]

    @validator("resource_type", always=True)
    def fix_resource_types(cls, v, values):
        """If the resource type does not match the unique_id prefix, then rewrite the resource type."""

        node_type = values.get("unique_id").split(".")[0]
        if v != node_type:
            return node_type
        return v

    @property
    def identifier(self) -> str:
        if not self.relation_name:
            return self.name

        return self.relation_name.split(".")[-1].replace('"', "").replace("`", "")

    def dump(self) -> Dict:
        """Dump the ManifestNode to a Dict, with support for pydantic 1 and 2"""
        exclude_set = {"schema_name", "depends_on", "node_config", "unique_id"}
        if hasattr(self, "model_dump"):
            return self.model_dump(exclude=exclude_set)  # type: ignore

        return self.dict(exclude=exclude_set)


class UnknownManifestPathType(Exception):
    """Raised when the ManifestLoader receives a FileReferenceConfig with a path that does not have a known URL scheme."""


class InvalidManifestPath(Exception):
    """Raised when the ManifestLoader receives a FileReferenceConfig with an invalid path."""


class ManifestLoader:
    def __init__(self):
        self.loading_functions = {
            ManifestReferenceType.file: self.load_from_path,
            ManifestReferenceType.dbt_cloud: self.load_from_dbt_cloud,
            ManifestReferenceType.gcs: self.load_from_gcs,
            ManifestReferenceType.s3: self.load_from_s3,
            ManifestReferenceType.azure: self.load_from_azure,
        }

    @staticmethod
    def load_from_path(config: FileReferenceConfig) -> Dict:
        """
        Load a manifest dictionary based on a FileReferenceConfig. This config's
        path can point to either a local file or a URL to a remote location.
        """

        if config.path.scheme in ("http", "https"):
            return ManifestLoader.load_from_http(config)

        if config.path.scheme in ("file"):
            return ManifestLoader.load_from_local_filesystem(config)

        raise UnknownManifestPathType()

    @staticmethod
    def load_from_local_filesystem(config: FileReferenceConfig) -> Dict:
        """Load a manifest dictionary from a local file"""

        if not config.path.path:
            raise InvalidManifestPath()

        file_path = Path(config.path.path)

        if not file_path.exists():
            raise LoomConfigurationError(f"The path `{file_path}` does not exist.")

        if file_path.suffix == ".gz":
            with gzip.open(file_path, "rt") as file:
                return json.load(file)

        return json.load(open(file_path))

    @staticmethod
    def load_from_http(config: FileReferenceConfig) -> Dict:
        """Load a manifest dictionary from a local file"""

        if not config.path.path:
            raise InvalidManifestPath()

        response = requests.get(urlunparse(config.path), stream=True)
        response.raise_for_status()  # Check for request errors

        # Check for compression on the file. If compressed, store it in a buffer
        # and decompress it.
        if (
            config.path.path.endswith(".gz")
            or response.headers.get("Content-Encoding") == "gzip"
        ):
            with gzip.GzipFile(fileobj=BytesIO(response.content)) as gz_file:
                return json.load(gz_file)

        return response.json()

    @staticmethod
    def load_from_dbt_cloud(config: DbtCloudReferenceConfig) -> Dict:
        """Load a manifest dictionary from dbt Cloud."""
        client = DbtCloud(
            account_id=config.account_id, api_endpoint=config.api_endpoint
        )

        return client.get_models(config.job_id, step=config.step)

    @staticmethod
    def load_from_gcs(config: GCSReferenceConfig) -> Dict:
        """Load a manifest dictionary from a GCS bucket."""
        gcs_client = GCSClient(
            project_id=config.project_id,
            bucket_name=config.bucket_name,
            object_name=config.object_name,
            credentials=config.credentials,
        )

        return gcs_client.load_manifest()

    @staticmethod
    def load_from_s3(config: S3ReferenceConfig) -> Dict:
        """Load a manifest dictionary from an S3-compatible bucket."""
        gcs_client = S3Client(
            bucket_name=config.bucket_name,
            object_name=config.object_name,
        )

        return gcs_client.load_manifest()

    @staticmethod
    def load_from_azure(config: AzureReferenceConfig) -> Dict:
        """Load a manifest dictionary from Azure storage."""
        azure_client = AzureClient(
            container_name=config.container_name,
            object_name=config.object_name,
            account_name=config.account_name,
        )

        return azure_client.load_manifest()

    def load(self, manifest_reference: ManifestReference) -> Dict:
        """Load a manifest dictionary based on a ManifestReference input."""

        if manifest_reference.type not in self.loading_functions:
            raise LoomConfigurationError(
                f"The manifest reference provided for {manifest_reference.name} does "
                "not have a valid type."
            )

        manifest = self.loading_functions[manifest_reference.type](
            manifest_reference.config
        )

        return manifest



from enum import Enum
from pathlib import Path
import re
from typing import List, Union
from urllib.parse import ParseResult, urlparse

from pydantic import BaseModel, Field, validator

from dbt_loom.clients.az_blob import AzureReferenceConfig
from dbt_loom.clients.dbt_cloud import DbtCloudReferenceConfig
from dbt_loom.clients.gcs import GCSReferenceConfig
from dbt_loom.clients.s3 import S3ReferenceConfig


class ManifestReferenceType(str, Enum):
    """Type of ManifestReference"""

    file = "file"
    dbt_cloud = "dbt_cloud"
    gcs = "gcs"
    s3 = "s3"
    azure = "azure"


class FileReferenceConfig(BaseModel):
    """Configuration for a file reference"""

    path: ParseResult

    @validator("path", pre=True, always=True)
    def default_path(cls, v, values) -> ParseResult:
        """
        Check if the provided path is a valid URL. If not, convert it into an
        absolute file path.
        """

        if isinstance(v, ParseResult):
            return v

        if bool(re.match(r"^[a-zA-Z][a-zA-Z0-9+.-]*://", v)):
            return urlparse(v)

        return urlparse("file://" + str(Path(v).absolute()))


class ManifestReference(BaseModel):
    """Reference information for a manifest to be loaded into dbt-loom."""

    name: str
    type: ManifestReferenceType
    config: Union[
        FileReferenceConfig,
        DbtCloudReferenceConfig,
        GCSReferenceConfig,
        S3ReferenceConfig,
        AzureReferenceConfig,
    ]
    excluded_packages: List[str] = Field(default_factory=list)


class dbtLoomConfig(BaseModel):
    """Configuration for dbt Loom"""

    manifests: List[ManifestReference]
    enable_telemetry: bool = False


class LoomConfigurationError(BaseException):
    """Error raised when dbt-loom has been misconfigured."""



from dataclasses import dataclass
import os
import re
from pathlib import Path
from typing import Callable, Dict, Optional, Set

import yaml
from dbt.contracts.graph.node_args import ModelNodeArgs
from dbt.contracts.graph.nodes import ModelNode

from dbt.plugins.manager import dbt_hook, dbtPlugin
from dbt.plugins.manifest import PluginNodes
from dbt.config.project import VarProvider

from dbt_loom.shims import is_invalid_private_ref, is_invalid_protected_ref

try:
    from dbt.artifacts.resources.types import NodeType
except ModuleNotFoundError:
    from dbt.node_types import NodeType  # type: ignore


from dbt_loom.config import dbtLoomConfig
from dbt_loom.logging import fire_event
from dbt_loom.manifests import ManifestLoader, ManifestNode

import importlib.metadata


@dataclass
class LoomModelNodeArgs(ModelNodeArgs):
    """A dbt-loom extension of ModelNodeArgs to preserve resource types across lineages."""

    resource_type: NodeType = NodeType.Model
    group: Optional[str] = None

    def __init__(self, **kwargs):
        super().__init__(
            **{
                key: value
                for key, value in kwargs.items()
                if key not in ("resource_type", "group")
            }
        )
        self.resource_type = kwargs.get("resource_type", NodeType.Model)
        self.group = kwargs.get("group")

    @property
    def unique_id(self) -> str:
        unique_id = f"{self.resource_type}.{self.package_name}.{self.name}"
        if self.version:
            unique_id = f"{unique_id}.v{self.version}"

        return unique_id


def identify_node_subgraph(manifest) -> Dict[str, ManifestNode]:
    """
    Identify all nodes that should be selected from the manifest, and return ManifestNodes.
    """

    output = {}

    # We're going to temporarily allow all nodes here.
    for unique_id in manifest["nodes"].keys():
        if unique_id.split(".")[0] in (NodeType.Test.value, NodeType.Macro.value):
            continue

        node = manifest.get("nodes", {}).get(unique_id)

        if not node:
            continue

        if node.get("access") is None:
            node["access"] = node.get("config", {}).get("access", "protected")

        # Versions may be floats or strings. Standardize on strings for compatibility.
        for key in ("version", "latest_version"):
            if node.get(key):
                node[key] = str(node[key])

        output[unique_id] = ManifestNode(**(node))

    return output


def convert_model_nodes_to_model_node_args(
    selected_nodes: Dict[str, ManifestNode],
) -> Dict[str, LoomModelNodeArgs]:
    """Generate a dictionary of ModelNodeArgs based on a dictionary of ModelNodes"""
    return {
        unique_id: LoomModelNodeArgs(
            schema=node.schema_name,
            identifier=node.identifier,
            **(node.dump()),
        )
        for unique_id, node in selected_nodes.items()
        if node is not None
    }


@dataclass
class LoomRunnableConfig:
    """A shim class to allow is_invalid_*_ref functions to correctly handle access for loom-injected models."""

    restrict_access: bool = True
    vars: VarProvider = VarProvider(vars={})


class dbtLoom(dbtPlugin):
    """
    dbtLoom is a dbt plugin that loads manifest files, parses a DAG from the manifest,
    and injects public nodes from imported manifest.
    """

    def __init__(self, project_name: str):
        # Log the version of dbt-loom being initialized
        fire_event(
            msg=f'Initializing dbt-loom={importlib.metadata.version("dbt-loom")}'
        )

        configuration_path = Path(
            os.environ.get("DBT_LOOM_CONFIG", "dbt_loom.config.yml")
        )

        self._manifest_loader = ManifestLoader()
        self.manifests: Dict[str, Dict] = {}

        self.config: Optional[dbtLoomConfig] = self.read_config(configuration_path)
        self.models: Dict[str, LoomModelNodeArgs] = {}

        self._patch_ref_protection()

        if not self.config or (self.config and not self.config.enable_telemetry):
            self._patch_plugin_telemetry()

        super().__init__(project_name)

    def _patch_ref_protection(self) -> None:
        """Patch out the ref protection functions for proper protections"""
        import dbt.contracts.graph.manifest

        fire_event(
            msg="dbt-loom: Patching ref protection methods to support dbt-loom dependencies."
        )

        dbt.contracts.graph.manifest.Manifest.is_invalid_protected_ref = (  # type: ignore
            self.dependency_wrapper(is_invalid_protected_ref)
        )
        dbt.contracts.graph.manifest.Manifest.is_invalid_private_ref = (  # type: ignore
            self.dependency_wrapper(is_invalid_private_ref)
        )

        dbt.parser.manifest.ManifestLoader.check_valid_group_config_node = (  # type: ignore
            self.group_validation_wrapper(
                dbt.parser.manifest.ManifestLoader.check_valid_group_config_node  # type: ignore
            )
        )

        dbt.contracts.graph.nodes.ModelNode.from_args = (  # type: ignore
            self.model_node_wrapper(dbt.contracts.graph.nodes.ModelNode.from_args)  # type: ignore
        )

    def _patch_plugin_telemetry(self) -> None:
        """Patch the plugin telemetry function to prevent tracking of dbt plugins."""
        import dbt.tracking

        dbt.tracking.track = self.tracking_wrapper(dbt.tracking.track)

    def tracking_wrapper(self, function) -> Callable:
        """Wrap the telemetry `track` function and return early if we're tracking plugin actions."""

        def outer_function(*args, **kwargs):
            """Check the context of the snowplow tracker message for references to loom. Return if present."""

            if any(
                [
                    self.__class__.__name__ in str(context_item.__dict__)
                    or "dbt-loom" in str(context_item.__dict__)
                    or "dbt_loom" in str(context_item.__dict__)
                    for context_item in kwargs.get("context", [])
                ]
            ):
                return

            return function(*args, **kwargs)

        return outer_function

    def model_node_wrapper(self, function) -> Callable:
        """Wrap the ModelNode.from_args function and inject extra properties from the LoomModelNodeArgs."""

        def outer_function(args: LoomModelNodeArgs) -> ModelNode:
            model = function(args)
            model.group = args.group
            return model

        return outer_function

    def group_validation_wrapper(self, function) -> Callable:
        """Wrap the check_valid_group_config_node function to inject upstream group names."""

        def outer_function(
            inner_self, groupable_node, valid_group_names: Set[str]
        ) -> bool:
            new_groups: Set[str] = {
                model.group for model in self.models.values() if model.group is not None
            }

            return function(
                inner_self, groupable_node, valid_group_names.union(new_groups)
            )

        return outer_function

    def dependency_wrapper(self, function) -> Callable:
        def outer_function(inner_self, node, target_model, dependencies) -> bool:
            if self.config is not None:
                for manifest_name in self.manifests.keys():
                    if manifest_name in dependencies:
                        continue

                    dependencies[manifest_name] = LoomRunnableConfig()

            return function(inner_self, node, target_model, dependencies)

        return outer_function

    def get_groups(self) -> Set[str]:
        """Get all groups defined in injected models."""

        return {
            model.group for model in self.models.values() if model.group is not None
        }

    def read_config(self, path: Path) -> Optional[dbtLoomConfig]:
        """Read the dbt-loom configuration file."""
        if not path.exists():
            return None

        with open(path) as file:
            config_content = file.read()

        config_content = self.replace_env_variables(config_content)

        return dbtLoomConfig(**yaml.load(config_content, yaml.SafeLoader))

    @staticmethod
    def replace_env_variables(config_str: str) -> str:
        """Replace environment variable placeholders in the configuration string."""
        pattern = r"\$(\w+)|\$\{([^}]+)\}"
        return re.sub(
            pattern,
            lambda match: os.environ.get(
                match.group(1) if match.group(1) is not None else match.group(2), ""
            ),
            config_str,
        )

    def initialize(self) -> None:
        """Initialize the plugin"""

        if self.models != {} or not self.config:
            return

        for manifest_reference in self.config.manifests:
            fire_event(
                msg=f"dbt-loom: Loading manifest for `{manifest_reference.name}`"
                f" from `{manifest_reference.type.value}`"
            )

            manifest = self._manifest_loader.load(manifest_reference)
            if manifest is None:
                continue

            # Find the official project name from the manifest metadata and use that as the manifests key.
            manifest_name = manifest.get("metadata", {}).get(
                "project_name", manifest_reference.name
            )
            self.manifests[manifest_name] = manifest

            selected_nodes = identify_node_subgraph(manifest)

            # Remove nodes from excluded packages.
            filtered_nodes = {
                key: value
                for key, value in selected_nodes.items()
                if value.package_name not in manifest_reference.excluded_packages
            }

            loom_nodes = convert_model_nodes_to_model_node_args(filtered_nodes)

            self.models.update(loom_nodes)

    @dbt_hook
    def get_nodes(self) -> PluginNodes:
        """
        Inject PluginNodes to dbt for injection into dbt's DAG.
        """
        fire_event(msg="dbt-loom: Injecting nodes")
        return PluginNodes(models=self.models)  # type: ignore


plugins = [dbtLoom]



try:
    import dbt_common.events.functions as dbt_event_function
    from dbt_common.events.types import Note
except ModuleNotFoundError:
    import dbt.events.functions as dbt_event_function  # type: ignore
    from dbt.events.types import Note  # type: ignore


def fire_event(*args, **kwargs) -> None:
    """Fire a dbt-core event."""
    dbt_event_function.fire_event(Note(*args, **kwargs))



import json
import os
import gzip
from io import BytesIO
from typing import Dict

from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobServiceClient
from pydantic import BaseModel


class AzureReferenceConfig(BaseModel):
    """Configuration for an reference stored in Azure Storage"""

    container_name: str
    object_name: str
    account_name: str


class AzureClient:
    """A client for loading manifest files from Azure storage."""

    def __init__(
        self, container_name: str, object_name: str, account_name: str
    ) -> None:
        self.account_name = account_name
        self.container_name = container_name
        self.object_name = object_name

    def load_manifest(self) -> Dict:
        """Load the manifest.json file from Azure storage."""

        connection_string = os.getenv("AZURE_STORAGE_CONNECTION_STRING")
        try:
            if connection_string:
                blob_service_client = BlobServiceClient.from_connection_string(
                    connection_string
                )
            else:
                account_url = f"{self.account_name}.blob.core.windows.net"
                blob_service_client = BlobServiceClient(
                    account_url, credential=DefaultAzureCredential()
                )
            blob_client = blob_service_client.get_blob_client(
                container=self.container_name, blob=self.object_name
            )
        except Exception as e:
            raise Exception(
                "Unable to connect to Azure. Please confirm your credentials, connection details, and network."
            )

        # Deserialize the body of the object.
        try:
            if self.object_name.endswith('.gz'):
                with gzip.GzipFile(fileobj=BytesIO(blob_client.download_blob().readall())) as gzipfile:
                    content = gzipfile.read().decode('utf-8')
            else:
                content = blob_client.download_blob(encoding="utf-8").readall()
        except Exception:
            raise Exception(
                f"Unable to read the data contained in the object `{self.object_name}"
            )

        try:
            return json.loads(content)
        except Exception:
            raise Exception(
                f"The object `{self.object_name}` does not contain valid JSON."
            )



from typing import Mapping, Optional
from dbt.contracts.graph.nodes import GraphMemberNode, ModelNode
from dbt.contracts.graph.manifest import MaybeNonSource

try:
    from dbt.artifacts.resources.types import NodeType, AccessType
except ModuleNotFoundError:
    from dbt.node_types import NodeType, AccessType  # type: ignore


def is_invalid_protected_ref(
    self,
    node: GraphMemberNode,
    target_model: MaybeNonSource,
    dependencies: Optional[Mapping],
) -> bool:
    dependencies = dependencies or {}
    if not isinstance(target_model, ModelNode):
        return False

    is_protected_ref = (
        target_model.access == AccessType.Protected
        # don't raise this reference error for ad hoc 'preview' queries
        and node.resource_type != NodeType.SqlOperation
        and node.resource_type != NodeType.RPCCall  # TODO: rm
    )
    target_dependency = dependencies.get(target_model.package_name)
    restrict_package_access = (
        target_dependency.restrict_access if target_dependency else False
    )

    return is_protected_ref and (
        node.package_name != target_model.package_name and restrict_package_access
    )


def is_invalid_private_ref(
    self,
    node: GraphMemberNode,
    target_model: MaybeNonSource,
    dependencies: Optional[Mapping],
) -> bool:
    dependencies = dependencies or {}
    if not isinstance(target_model, ModelNode):
        return False

    is_private_ref = (
        target_model.access == AccessType.Private
        # don't raise this reference error for ad hoc 'preview' queries
        and node.resource_type != NodeType.SqlOperation
        and node.resource_type != NodeType.RPCCall  # TODO: rm
    )
    target_dependency = dependencies.get(target_model.package_name)
    restrict_package_access = (
        target_dependency.restrict_access if target_dependency else False
    )

    return is_private_ref and (
        # Invalid reference because the group does not match
        (hasattr(node, "group") and node.group and node.group != target_model.group)  # type: ignore
        # Or, invalid because these are different namespaces (project/package) and restrict-access is enforced
        or (node.package_name != target_model.package_name and restrict_package_access)
    )
